from tests.unit import UnitTestBase
from express.properties.scalar.scalar_property_context import ScalarPropertyFromContext

VALENCE_BAND_OFFSET = {
    "units": "eV",
    "name": "valence_band_offset",
    "value": 1
}


class ValenceBandOffsetTest(UnitTestBase):
    def setUp(self):
        super().setUp()

    def tearDown(self):
        super().tearDown()

    def test_valence_band_offset(self):
        parser = self.get_mocked_parser("valence_band_offset", 1)
        property_ = ScalarPropertyFromContext("valence_band_offset", None, value=1)
        self.assertDeepAlmostEqual(property_.serialize_and_validate(), VALENCE_BAND_OFFSET)



import os
import numpy as np

from express.parsers import BaseParser
from express.parsers.settings import Constant
from express.parsers.apps.espresso import settings
from express.parsers.mixins.ionic import IonicDataMixin
from express.parsers.mixins.reciprocal import ReciprocalDataMixin
from express.parsers.mixins.electronic import ElectronicDataMixin
from express.parsers.utils import find_file, lattice_basis_to_poscar
from express.parsers.apps.espresso.settings import NEB_PATH_FILE_SUFFIX
from express.parsers.apps.espresso.formats.txt import EspressoTXTParser
from express.parsers.apps.espresso.formats.xml import EspressoXMLParser


class EspressoParser(BaseParser, IonicDataMixin, ElectronicDataMixin, ReciprocalDataMixin):
    """
    Espresso parser class.
    """

    def __init__(self, *args, **kwargs):
        super(EspressoParser, self).__init__(*args, **kwargs)
        self.work_dir = self.kwargs["work_dir"]
        self.stdout_file = self.kwargs["stdout_file"]
        self.txt_parser = EspressoTXTParser(self.work_dir)
        self.xml_parser = EspressoXMLParser(self.find_xml_file())

    def find_xml_file(self):
        """
        Finds XML file.

        Note: Sternheimer GW XML file of the first process (gw0) is returned if this a Sternheimer GW calculation.

        Returns:
            str
        """
        is_sternheimer_gw = self._is_sternheimer_gw_calculation()
        for root, dirs, files in os.walk(self.work_dir, followlinks=True):
            for file_ in [f for f in files if settings.XML_DATA_FILE == f]:
                file_path = os.path.join(root, file_)
                if not is_sternheimer_gw or (is_sternheimer_gw and settings.STERNHEIMER_GW0_DIR_PATTERN in file_path):
                    return file_path

    def total_energy(self):
        """
        Returns total energy.

        Reference:
            func: express.parsers.mixins.electronic.ElectronicDataMixin.total_energy
        """
        return self.txt_parser.total_energy(self._get_file_content(self.stdout_file))

    def fermi_energy(self):
        """
        Returns fermi energy.

        Reference:
            func: express.parsers.mixins.electronic.ElectronicDataMixin.fermi_energy
        """
        return self.xml_parser.fermi_energy()

    def nspins(self):
        """
        Returns the number of spins.

        Reference:
            func: express.parsers.mixins.electronic.ElectronicDataMixin.nspins
        """
        return self.xml_parser.nspins()

    def _is_sternheimer_gw_calculation(self):
        """
        Checks whether this is a Sternheimer GW calculation.

        The calculation is considered Sternheimer GW if "SternheimerGW" is written on top of the stdout file.

        NOTE: DO NOT READ THE WHOLE FILE INTO MEMORY AS IT IS BIG.

        Returns:
             bool
        """
        if os.path.exists(self.stdout_file):
            with open(self.stdout_file, "r") as f:
                for index, line in enumerate(f):
                    if index > 50: return False
                    if settings.STERNHEIMER_GW_TITLE in line:
                        return True

    def eigenvalues_at_kpoints(self):
        """
        Returns eigenvalues for all kpoints.

        NOTE: eigenvalues are extracted from Sternheimer GW stdout file if this is a Sternheimer GW calculation.

        Reference:
            func: express.parsers.mixins.electronic.ElectronicDataMixin.eigenvalues_at_kpoints
        """
        if self._is_sternheimer_gw_calculation():
            text = self._get_file_content(self.stdout_file)
            inverse_reciprocal_lattice_vectors = self.xml_parser.get_inverse_reciprocal_lattice_vectors()
            return self.txt_parser.eigenvalues_at_kpoints_from_sternheimer_gw_stdout(text,
                                                                                     inverse_reciprocal_lattice_vectors)
        else:
            return self.xml_parser.eigenvalues_at_kpoints()

    def ibz_k_points(self):
        """
        Returns ibz_k_points.

        Note:
            The function assumes that kpoints extracted from parsed source are inside the irreducible wedge of the
            Brillouin zone. Without checking whether it is the case or not.

        Reference:
            func: express.parsers.mixins.reciprocal.ReciprocalDataMixin.ibz_k_points
        """
        return np.array([eigenvalueData["kpoint"] for eigenvalueData in self.eigenvalues_at_kpoints()])

    def dos(self):
        """
        Returns density of states.

        Reference:
            func: express.parsers.mixins.electronic.ElectronicDataMixin.dos
        """
        return self.txt_parser.dos()

    def initial_basis(self):
        """
        Returns initial basis.

        Reference:
            func: express.parsers.mixins.ionic.IonicDataMixin.initial_basis
        """
        return self.txt_parser.initial_basis(self._get_file_content(self.stdout_file))

    def initial_lattice_vectors(self):
        """
        Returns initial lattice vectors.

        Reference:
            func: express.parsers.mixins.ionic.IonicDataMixin.initial_lattice_vectors
        """
        return self.txt_parser.initial_lattice_vectors(self._get_file_content(self.stdout_file))

    def final_basis(self):
        """
        Returns final basis.

        Reference:
            func: express.parsers.mixins.ionic.IonicDataMixin.basis
        """
        return self.xml_parser.final_basis()

    def final_lattice_vectors(self):
        """
        Returns final lattice.

        Reference:
            func: express.parsers.mixins.ionic.IonicDataMixin.lattice_vectors
        """
        return self.xml_parser.final_lattice_vectors(reciprocal=False)

    def convergence_electronic(self):
        """
        Extracts convergence electronic.

        Reference:
            func: express.parsers.mixins.electronic.ElectronicDataMixin.convergence_electronic
        """
        return self.txt_parser.convergence_electronic(self._get_file_content(self.stdout_file))

    def convergence_ionic(self):
        """
        Returns convergence ionic.

        Reference:
            func: express.parsers.mixins.ionic.IonicDataMixin.convergence_ionic
        """
        return self.txt_parser.convergence_ionic(self._get_file_content(self.stdout_file))

    def stress_tensor(self):
        """
        Returns stress tensor.

        Reference:
            func: express.parsers.mixins.ionic.IonicDataMixin.stress_tensor
        """
        return self.txt_parser.stress_tensor(self._get_file_content(self.stdout_file))

    def pressure(self):
        """
        Returns pressure.

        Reference:
            func: express.parsers.mixins.ionic.IonicDataMixin.pressure
        """
        return self.txt_parser.pressure(self._get_file_content(self.stdout_file))

    def total_force(self):
        """
        Returns total force.

        Reference:
            func: express.parsers.mixins.ionic.IonicDataMixin.total_force
        """
        return self.txt_parser.total_force(self._get_file_content(self.stdout_file))

    def atomic_forces(self):
        """
        Returns forces that is exerted on each atom by its surroundings.

        Reference:
            func: express.parsers.mixins.ionic.IonicDataMixin.atomic_forces
        """
        return self.txt_parser.atomic_forces(self._get_file_content(self.stdout_file))

    def total_energy_contributions(self):
        """
        Extracts total energy contributions.

        Reference:
            func: express.parsers.mixins.electronic.ElectronicDataMixin.total_energy_contributions
        """
        return self.txt_parser.total_energy_contributions(self._get_file_content(self.stdout_file))

    def zero_point_energy(self):
        """
        Returns zero point energy.

        Reference:
            func: express.parsers.mixins.ionic.IonicDataMixin.zero_point_energy
        """
        return self.txt_parser.zero_point_energy(self._get_file_content(self.stdout_file))

    def phonon_dos(self):
        """
        Returns phonon dos.

        Reference:
            func: express.parsers.mixins.ionic.IonicDataMixin.phonon_dos
        """
        return self.txt_parser.phonon_dos()

    def phonon_dispersions(self):
        """
        Returns phonon dispersions.

        Reference:
            func: express.parsers.mixins.ionic.IonicDataMixin.phonon_dispersions
        """
        return self.txt_parser.phonon_dispersions()

    def _find_neb_dat_file(self):
        neb_path_file = find_file(NEB_PATH_FILE_SUFFIX, self.work_dir)
        if neb_path_file: return "{}.dat".format(neb_path_file[:neb_path_file.rfind(".")])

    def reaction_coordinates(self):
        """
        Returns reaction coordinates.

        Returns:
             list
        """
        neb_dat_file = self._find_neb_dat_file()
        return self.txt_parser.reaction_coordinates(self._get_file_content(neb_dat_file))

    def reaction_energies(self):
        """
        Returns reaction energies.

        Returns:
             list
        """
        neb_dat_file = self._find_neb_dat_file()
        return self.txt_parser.reaction_energies(self._get_file_content(neb_dat_file))

    def _get_esm_file(self):
        return find_file(".esm1", self.work_dir)

    def potential_profile(self):
        return self.txt_parser.potential_profile(self._get_file_content(self._get_esm_file()))

    def charge_density_profile(self):
        return self.txt_parser.charge_density_profile(self._get_file_content(self._get_esm_file()))

    def _is_pw_scf_output_file(self, path):
        """
        Checks whether the given file is PWSCF output file.

        The file is considered PWSCF output file if "Program PWSCF" is written on top of the file.

        NOTE: DO NOT READ THE WHOLE FILE INTO MEMORY AS IT COULD BE BIG.

        Returns:
             bool
        """
        if os.path.exists(path):
            with open(path, "r") as f:
                for index, line in enumerate(f):
                    if index > 50: return False
                    if settings.PWSCF_OUTPUT_FILE_REGEX in line:
                        return True

    def _find_pw_scf_output_files(self):
        pw_scf_output_files = []
        for root, dirs, files in os.walk(self.work_dir, followlinks=True):
            for file in files:
                path = os.path.join(root, file)
                if self._is_pw_scf_output_file(path):
                    pw_scf_output_files.append(path)
        return pw_scf_output_files

    def initial_structure_strings(self):
        structures = []
        for pw_scf_output_file in self._find_pw_scf_output_files():
            try:
                basis = self.txt_parser.initial_basis(self._get_file_content(pw_scf_output_file))
                lattice = self.txt_parser.initial_lattice_vectors(self._get_file_content(pw_scf_output_file))
                structures.append(lattice_basis_to_poscar(lattice, basis))
            except:
                raise
        return structures

    def final_structure_strings(self):
        structures = []
        for pw_scf_output_file in self._find_pw_scf_output_files():
            try:
                basis = self.txt_parser.final_basis(self._get_file_content(pw_scf_output_file))
                lattice = self.txt_parser.final_lattice_vectors(self._get_file_content(pw_scf_output_file))
                structures.append(lattice_basis_to_poscar(lattice, basis))
            except:
                pass
        return structures

    def averaged_potential(self):
        data = self.txt_parser.averaged_quantity(self.stdout_file)
        data["x"] *= Constant.BOHR  # convert to angstrom
        data["planar_average"] *= Constant.RYDBERG  # convert to eV
        data["macroscopic_average"] *= Constant.RYDBERG  # convert to eV
        return data



from express.properties.non_scalar.two_dimensional_plot import TwoDimensionalPlotProperty


class AveragedPotentialProfile(TwoDimensionalPlotProperty):
    """
    Averaged potential profile plot.
    """

    def __init__(self, name: str, parser, *args, **kwargs):
        super().__init__(name, parser, *args, **kwargs)

        self.data = self.safely_invoke_parser_method("averaged_potential")
        self.xDataArray = self.data["x"].tolist()
        self.yDataSeries = [self.data["planar_average"].tolist(), self.data["macroscopic_average"].tolist()]



from typing import Type, Any
from express.properties.scalar import ScalarProperty
from express.parsers import BaseParser


class ScalarPropertyFromContext(ScalarProperty):

    def __init__(self,
                 name: str,
                 parser: Type[BaseParser],
                 value: Any,
                 *args,
                 **kwargs):
        super().__init__(name, parser, *args, **kwargs)
        self.value = value



import numpy as np
from typing import Tuple
from copy import deepcopy
from express import settings
from express.properties.utils import eigenvalues
from express.properties.non_scalar import NonScalarProperty


class BandGaps(NonScalarProperty):
    """
    The minimum energy difference between the highest occupied (valence) band and the lowest unoccupied ( conduction)
    band, extracted from the k-mesh of electronic eigenvalues (default) or the bandstructure. Can be direct - when
    the difference is between the energy points at the same point in reciprocal space, and indirect - when the
    difference between two inequivalent points. The flavor defines the type of band_gap computation. It could be done
    either using 'mesh' or 'path'.
    """

    def __init__(self, name, parser, *args, **kwargs):
        super(BandGaps, self).__init__(name, parser, *args, **kwargs)

        self.values = None
        self.nspins = self.safely_invoke_parser_method("nspins")
        self.ibz_k_points = self.safely_invoke_parser_method("ibz_k_points")
        self.fermi_energy = self.safely_invoke_parser_method("fermi_energy")
        self.band_gaps_direct = self.safely_invoke_parser_method("band_gaps_direct")
        self.band_gaps_indirect = self.safely_invoke_parser_method("band_gaps_indirect")
        self.eigenvalues_at_kpoints = self.safely_invoke_parser_method("eigenvalues_at_kpoints")

        if self.band_gaps_direct is not None and self.band_gaps_indirect is not None:
            self.values = [
                self._serialize_band_gaps(self.band_gaps_direct, "direct"),
                self._serialize_band_gaps(self.band_gaps_indirect, "indirect")
            ]

    def _serialize(self) -> dict:
        return {
            "name": self.name,
            "values": self.values if self.values else self.get_band_gaps_from_mesh(),
            "eigenvalues": self._eigenvalues() if not self.values else []
        }

    def _serialize_band_gaps(self, gap: float, gap_type: str, spin: float = 1/2) -> dict:
        return {
            "type": gap_type,
            "units": self.manifest["defaults"]["units"],
            "value": gap,
            "spin": spin,
        }

    def get_band_gaps_from_mesh(self) -> list:
        """
        Compute direct and indirect band gaps for all available spins
        """
        occupations, eigenvalue_mesh = self._get_bands_info()

        gap_types = ["direct", "indirect"]
        computed_gaps = []

        for s in range(self.nspins):
            for gap_type in gap_types:
                computed_gaps.append(
                    self.compute_on_mesh(
                        eigenvalue_mesh=eigenvalue_mesh,
                        occupations=occupations,
                        spin_channel=s,
                        gap_type=gap_type
                    )
                )
        return computed_gaps

    def compute_on_mesh(self,
                        eigenvalue_mesh: np.ndarray,
                        occupations: np.ndarray,
                        spin_channel: int = 0,
                        gap_type: str = "indirect",
                        absolute_eigenvalues: bool = True) -> dict:
        """
        Calculates the band gap on the material's mesh for a given gap type and spin channel.

        Args:
            eigenvalue_mesh:      3D array of eigenvalues with indices in following order: spin, kpoint, eigenvalue
            occupations:          2D array of number of occupied bands per spin and k-point (order: spin, kpoint)
            spin_channel:         spin index (0 -> 1/2, 1 -> -1/2)
            gap_type:             band gap type, either direct or indirect
            absolute_eigenvalues: whether to unshift the eigenvalues (applies to band edge values only)

        Returns:
            band gap dictionary
        """
        ev_k = eigenvalue_mesh[spin_channel, :, 0]  # valence band of current spin channel
        ec_k = eigenvalue_mesh[spin_channel, :, 1]  # conduction band of current spin channel
        occ_k = occupations[spin_channel]           # band occupations for current spin channel
        spin = 1/2*(-1)**spin_channel               # spin value

        gap, k_val, k_cond = BandGaps._find_gap(occ_k, ev_k, ec_k, gap_type=gap_type)
        result = self._serialize_band_gaps(gap=gap, gap_type=gap_type, spin=spin)

        # value for shifting back eigenvalues (see also _get_bands_info)
        e_fermi = self.fermi_energy if absolute_eigenvalues else 0

        if k_val is not None and k_cond is not None:
            result.update({
                "kpointValence": self._round(self.ibz_k_points[k_val]),
                "kpointConduction": self._round(self.ibz_k_points[k_cond]),
                "eigenvalueValence": ev_k[k_val] + e_fermi,
                "eigenvalueConduction": ec_k[k_cond] + e_fermi,
            })
        return result

    def _get_bands_info(self) -> Tuple[np.ndarray, np.ndarray]:
        """
        Extracts bands information:
            - number of kpoints (nk) - int
            - number of spins (ns) - int
            - eigenvalue mesh (e_skn) of valence and conduction band shifted by Fermi energy - 3D array
            - number of occupied bands per spin and k-point (occ_sk) - 2D array

        Note: spin indices correspond to spin value as follows (see also utils.eigenvalues function)
              index       value
              s = 0  <->  1/2
              s = 1  <-> -1/2

        Returns:
            tuple: bands information containing e_skn and occ_sk explained above.

        """
        nk = len(self.ibz_k_points)
        ns = self.nspins
        e_skn = np.array([[eigenvalues(self.eigenvalues_at_kpoints, k, s) for k in range(nk)] for s in range(ns)])
        e_skn -= self.fermi_energy
        occ_sk = (e_skn < 0.0).sum(2)
        # select highest occupied and lowest unoccupied bands
        e_skn = np.array([[e_skn[s, k, occ_sk[s, k] - 1:occ_sk[s, k] + 1] for k in range(nk)] for s in range(ns)])
        return occ_sk, e_skn

    @staticmethod
    def _find_gap(occupations: np.ndarray,
                  valence_band: np.ndarray,
                  conduction_band: np.ndarray,
                  gap_type: str = "indirect") -> Tuple[float, int, int]:
        """
        Computes the difference in energy between the highest valence band and the lowest conduction band.

        Args:
            occupations:     numbers of occupied bands per k-point.
            valence_band:    eigenvalues of valence band per k-point.
            conduction_band: eiganvalues of conduction band per k-point.
            gap_type:        band gap type, either direct or indirect.

        Returns:
            tuple: a (gap, k1, k2) tuple where k1 and k2 are the indices of the valence and conduction k-points.
        """
        if occupations.ptp() > 0:
            # Some band must be crossing fermi-level. Hence, we return zero for band gap and the actual k-points
            kv = kc = occupations.argmax()
            return 0.0, kv, kc
        if gap_type == "direct":
            direct_gaps = conduction_band - valence_band
            k = direct_gaps.argmin()
            return direct_gaps[k], k, k
        kv = valence_band.argmax()
        kc = conduction_band.argmin()
        return conduction_band[kc] - valence_band[kv], kv, kc

    def _eigenvalues(self) -> list:
        """
        Extracts eigenvalues around Fermi level, i.e. last two values in occupation 1 and first two values in occupation 0.

        Returns:
             dict
        """
        eigens_at_kpoints = deepcopy(self.eigenvalues_at_kpoints)
        for eigens_at_kpoint in eigens_at_kpoints:
            eigens_at_kpoint["kpoint"] = self._round(eigens_at_kpoint["kpoint"])
            for eigens_at_spin in eigens_at_kpoint["eigenvalues"]:
                eigens_at_spin["energies"] = self._round(eigens_at_spin["energies"])
                eigens_at_spin["occupations"] = self._round(eigens_at_spin["occupations"])
                # occupations are empty in case of QE GW, hence sending all values.
                if len(eigens_at_spin["occupations"]) == 0:
                    continue
                start = max(0, len(eigens_at_spin["occupations"]) - eigens_at_spin["occupations"][::-1].index(1.0) - 2)
                end = min(len(eigens_at_spin["occupations"]), eigens_at_spin["occupations"].index(0.0) + 2)
                eigens_at_spin["energies"] = eigens_at_spin["energies"][start:end]
                eigens_at_spin["occupations"] = eigens_at_spin["occupations"][start:end]
        return eigens_at_kpoints

    def _round(self, array):
        return np.round(array, settings.PRECISION).tolist()



from mock import MagicMock
import numpy as np
from tests.unit import UnitTestBase
from express.properties.non_scalar.two_dimensional_plot.averaged_potential_profile import AveragedPotentialProfile


AVG_POTENTIAL = {
    "name": "averaged_potential_profile",
    "xDataArray": [0, 2.0943951, 4.1887902, 6.28318531],
    "yDataSeries": [[0, 0, 0, 0], [1.0, -0.5, -0.5, 1.0]],
    "yAxis": {
        "label": "energy",
        "units": "eV"
    },
    "xAxis": {
        "label": "z coordinate",
        "units": "angstrom"
    },
}


class AveragedPotentialProfileTest(UnitTestBase):
    def setUp(self):
        super().setUp()

    def tearDown(self):
        super().tearDown()

    def create_mock_data(self, n_points: int = 300):
        """
        Creates mock potential data based on cosine in the range [0, 2*pi].
        Exact minimum is at
          x                   = pi
          macroscopic_average = -1
        """
        x = np.linspace(0, 2*np.pi, n_points)
        p_x = np.zeros(n_points)  # ignored for now
        m_x = np.cos(x)
        dtype = np.dtype([("x", float), ("planar_average", float), ("macroscopic_average", float)])
        return np.array(list(zip(x, p_x, m_x)), dtype=dtype)

    def test_averaged_potential(self):
        averaged_potential_data = self.create_mock_data(n_points=4)
        parser = MagicMock()
        parser.attach_mock(MagicMock(return_value=averaged_potential_data), "averaged_potential")
        property_ = AveragedPotentialProfile("averaged_potential_profile", parser)
        self.assertDeepAlmostEqual(property_.serialize_and_validate(), AVG_POTENTIAL, places=6)
